@@ -10,6 +10,8 @@ import (
 	"syscall"
 	"time"
 
+	"golang.org/x/sys/unix"
+
 	"github.com/hanwen/go-fuse/fuse"
 	"github.com/hanwen/go-fuse/fuse/nodefs"
 	"github.com/hanwen/go-fuse/fuse/pathfs"
@@ -193,7 +195,7 @@ func (fs *FS) Create(path string, flags uint32, mode uint32, context *fuse.Conte
 	cName := filepath.Base(cPath)
 
 	// Handle long file name
-	if nametransform.IsLongContent(cName) {
+	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
 		var dirfd *os.File
 		dirfd, err = os.Open(filepath.Dir(cPath))
 		if err != nil {
@@ -209,7 +211,7 @@ func (fs *FS) Create(path string, flags uint32, mode uint32, context *fuse.Conte
 
 		// Create content
 		var fdRaw int
-		fdRaw, err = syscallcompat.Openat(int(dirfd.Fd()), cName, newFlags|os.O_CREATE, mode)
+		fdRaw, err = syscallcompat.Openat(int(dirfd.Fd()), cName, newFlags|os.O_CREATE|os.O_EXCL, mode)
 		if err != nil {
 			nametransform.DeleteLongName(dirfd, cName)
 			return nil, fuse.ToStatus(err)
@@ -217,7 +219,7 @@ func (fs *FS) Create(path string, flags uint32, mode uint32, context *fuse.Conte
 		fd = os.NewFile(uintptr(fdRaw), cName)
 	} else {
 		// Normal (short) file name
-		fd, err = os.OpenFile(cPath, newFlags|os.O_CREATE, os.FileMode(mode))
+		fd, err = os.OpenFile(cPath, newFlags|os.O_CREATE|os.O_EXCL, os.FileMode(mode))
 		if err != nil {
 			return nil, fuse.ToStatus(err)
 		}
@@ -237,13 +239,14 @@ func (fs *FS) Chmod(path string, mode uint32, context *fuse.Context) (code fuse.
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.getBackingPath(path)
+	dirfd, cName, err := fs.openBackingPath(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
+	defer dirfd.Close()
 	// os.Chmod goes through the "syscallMode" translation function that messes
-	// up the suid and sgid bits. So use syscall.Chmod directly.
-	err = syscall.Chmod(cPath, mode)
+	// up the suid and sgid bits. So use a syscall directly.
+	err = syscallcompat.Fchmodat(int(dirfd.Fd()), cName, mode, unix.AT_SYMLINK_NOFOLLOW)
 	return fuse.ToStatus(err)
 }
 
@@ -252,21 +255,22 @@ func (fs *FS) Chown(path string, uid uint32, gid uint32, context *fuse.Context)
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.getBackingPath(path)
+	dirfd, cName, err := fs.openBackingPath(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	code = fuse.ToStatus(os.Lchown(cPath, int(uid), int(gid)))
+	defer dirfd.Close()
+	code = fuse.ToStatus(syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(uid), int(gid), unix.AT_SYMLINK_NOFOLLOW))
 	if !code.Ok() {
 		return code
 	}
 	if !fs.args.PlaintextNames {
 		// When filename encryption is active, every directory contains
 		// a "gocryptfs.diriv" file. This file should also change the owner.
-		// Instead of checking if "cPath" is a directory, we just blindly
-		// execute the Lchown on "cPath/gocryptfs.diriv" and ignore errors.
-		dirIVPath := filepath.Join(cPath, nametransform.DirIVFilename)
-		os.Lchown(dirIVPath, int(uid), int(gid))
+		// Instead of checking if "cName" is a directory, we just blindly
+		// execute the chown on "cName/gocryptfs.diriv" and ignore errors.
+		dirIVPath := filepath.Join(cName, nametransform.DirIVFilename)
+		syscallcompat.Fchownat(int(dirfd.Fd()), dirIVPath, int(uid), int(gid), unix.AT_SYMLINK_NOFOLLOW)
 	}
 	return fuse.OK
 }
@@ -276,19 +280,13 @@ func (fs *FS) Mknod(path string, mode uint32, dev uint32, context *fuse.Context)
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.getBackingPath(path)
+	dirfd, cName, err := fs.openBackingPath(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	// Create ".name" file to store long file name
-	cName := filepath.Base(cPath)
-	if nametransform.IsLongContent(cName) {
-		var dirfd *os.File
-		dirfd, err = os.Open(filepath.Dir(cPath))
-		if err != nil {
-			return fuse.ToStatus(err)
-		}
-		defer dirfd.Close()
+	defer dirfd.Close()
+	// Create ".name" file to store long file name (except in PlaintextNames mode)
+	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
 		err = fs.nameTransform.WriteLongName(dirfd, cName, path)
 		if err != nil {
 			return fuse.ToStatus(err)
@@ -300,16 +298,17 @@ func (fs *FS) Mknod(path string, mode uint32, dev uint32, context *fuse.Context)
 		}
 	} else {
 		// Create regular device node
-		err = syscall.Mknod(cPath, mode, int(dev))
+		err = syscallcompat.Mknodat(int(dirfd.Fd()), cName, mode, int(dev))
 	}
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
 	// Set owner
 	if fs.args.PreserveOwner {
-		err = os.Lchown(cPath, int(context.Owner.Uid), int(context.Owner.Gid))
+		err = syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(context.Owner.Uid),
+			int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
 		if err != nil {
-			tlog.Warn.Printf("Mknod: Lchown failed: %v", err)
+			tlog.Warn.Printf("Mknod: Fchownat failed: %v", err)
 		}
 	}
 	return fuse.OK
@@ -385,33 +384,23 @@ func (fs *FS) Unlink(path string, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.getBackingPath(path)
+	dirfd, cName, err := fs.openBackingPath(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-
-	cName := filepath.Base(cPath)
-	if nametransform.IsLongContent(cName) {
-		var dirfd *os.File
-		dirfd, err = os.Open(filepath.Dir(cPath))
-		if err != nil {
-			return fuse.ToStatus(err)
-		}
-		defer dirfd.Close()
-		// Delete content
-		err = syscallcompat.Unlinkat(int(dirfd.Fd()), cName)
-		if err != nil {
-			return fuse.ToStatus(err)
-		}
-		// Delete ".name"
+	defer dirfd.Close()
+	// Delete content
+	err = syscallcompat.Unlinkat(int(dirfd.Fd()), cName, 0)
+	if err != nil {
+		return fuse.ToStatus(err)
+	}
+	// Delete ".name" file
+	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
 		err = nametransform.DeleteLongName(dirfd, cName)
 		if err != nil {
 			tlog.Warn.Printf("Unlink: could not delete .name file: %v", err)
 		}
-		return fuse.ToStatus(err)
 	}
-
-	err = syscall.Unlink(cPath)
 	return fuse.ToStatus(err)
 }
 
@@ -421,49 +410,41 @@ func (fs *FS) Symlink(target string, linkName string, context *fuse.Context) (co
 	if fs.isFiltered(linkName) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.getBackingPath(linkName)
+	dirfd, cName, err := fs.openBackingPath(linkName)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	if fs.args.PlaintextNames {
-		err = os.Symlink(target, cPath)
-		return fuse.ToStatus(err)
+	defer dirfd.Close()
+	var cTarget string = target
+	if !fs.args.PlaintextNames {
+		// Symlinks are encrypted like file contents (GCM) and base64-encoded
+		cBinTarget := fs.contentEnc.EncryptBlock([]byte(target), 0, nil)
+		cTarget = fs.nameTransform.B64.EncodeToString(cBinTarget)
 	}
-	// Symlinks are encrypted like file contents (GCM) and base64-encoded
-	cBinTarget := fs.contentEnc.EncryptBlock([]byte(target), 0, nil)
-	cTarget := fs.nameTransform.B64.EncodeToString(cBinTarget)
-	// Handle long file name
-	cName := filepath.Base(cPath)
-	if nametransform.IsLongContent(cName) {
-		var dirfd *os.File
-		dirfd, err = os.Open(filepath.Dir(cPath))
-		if err != nil {
-			return fuse.ToStatus(err)
-		}
-		defer dirfd.Close()
-		// Create ".name" file
+	// Create ".name" file to store long file name (except in PlaintextNames mode)
+	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
 		err = fs.nameTransform.WriteLongName(dirfd, cName, linkName)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
 		// Create "gocryptfs.longfile." symlink
-		// TODO use syscall.Symlinkat once it is available in Go
-		err = syscall.Symlink(cTarget, cPath)
+		err = syscallcompat.Symlinkat(cTarget, int(dirfd.Fd()), cName)
 		if err != nil {
 			nametransform.DeleteLongName(dirfd, cName)
 		}
 	} else {
 		// Create symlink
-		err = os.Symlink(cTarget, cPath)
+		err = syscallcompat.Symlinkat(cTarget, int(dirfd.Fd()), cName)
 	}
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
 	// Set owner
 	if fs.args.PreserveOwner {
-		err = os.Lchown(cPath, int(context.Owner.Uid), int(context.Owner.Gid))
+		err = syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(context.Owner.Uid),
+			int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
 		if err != nil {
-			tlog.Warn.Printf("Mknod: Lchown failed: %v", err)
+			tlog.Warn.Printf("Symlink: Fchownat failed: %v", err)
 		}
 	}
 	return fuse.OK
@@ -485,7 +466,10 @@ func (fs *FS) Rename(oldPath string, newPath string, context *fuse.Context) (cod
 	// The Rename may cause a directory to take the place of another directory.
 	// That directory may still be in the DirIV cache, clear it.
 	fs.nameTransform.DirIVCache.Clear()
-
+	// Easy case.
+	if fs.args.PlaintextNames {
+		return fuse.ToStatus(syscall.Rename(cOldPath, cNewPath))
+	}
 	// Handle long source file name
 	var oldDirFd *os.File
 	var finalOldDirFd int
@@ -517,7 +501,13 @@ func (fs *FS) Rename(oldPath string, newPath string, context *fuse.Context) (cod
 		finalNewPath = cNewName
 		// Create destination .name file
 		err = fs.nameTransform.WriteLongName(newDirFd, cNewName, newPath)
-		if err != nil {
+		// Failure to write the .name file is expected when the target path already
+		// exists. Since hashes are pretty unique, there is no need to modify the
+		// file anyway. We still set newDirFd to nil to ensure that we do not delete
+		// the file on error.
+		if err == syscall.EEXIST {
+			newDirFd = nil
+		} else if err != nil {
 			return fuse.ToStatus(err)
 		}
 	}
@@ -553,35 +543,32 @@ func (fs *FS) Link(oldPath string, newPath string, context *fuse.Context) (code
 	if fs.isFiltered(newPath) {
 		return fuse.EPERM
 	}
-	cOldPath, err := fs.getBackingPath(oldPath)
+	oldDirFd, cOldName, err := fs.openBackingPath(oldPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	cNewPath, err := fs.getBackingPath(newPath)
+	defer oldDirFd.Close()
+	newDirFd, cNewName, err := fs.openBackingPath(newPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	// Handle long file name
-	cNewName := filepath.Base(cNewPath)
-	if nametransform.IsLongContent(cNewName) {
-		dirfd, err := os.Open(filepath.Dir(cNewPath))
+	defer newDirFd.Close()
+	// Handle long file name (except in PlaintextNames mode)
+	if !fs.args.PlaintextNames && nametransform.IsLongContent(cNewName) {
+		err = fs.nameTransform.WriteLongName(newDirFd, cNewName, newPath)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
-		defer dirfd.Close()
-		err = fs.nameTransform.WriteLongName(dirfd, cNewName, newPath)
+		// Create "gocryptfs.longfile." link
+		err = syscallcompat.Linkat(int(oldDirFd.Fd()), cOldName, int(newDirFd.Fd()), cNewName, 0)
 		if err != nil {
-			return fuse.ToStatus(err)
+			nametransform.DeleteLongName(newDirFd, cNewName)
 		}
-		// TODO Use syscall.Linkat once it is available in Go (it is not in Go
-		// 1.6).
-		err = syscall.Link(cOldPath, cNewPath)
-		if err != nil {
-			nametransform.DeleteLongName(dirfd, cNewName)
-		}
-		return fuse.ToStatus(err)
+	} else {
+		// Create regular link
+		err = syscallcompat.Linkat(int(oldDirFd.Fd()), cOldName, int(newDirFd.Fd()), cNewName, 0)
 	}
-	return fuse.ToStatus(os.Link(cOldPath, cNewPath))
+	return fuse.ToStatus(err)
 }
 
 // Access implements pathfs.Filesystem.
